/*
 * Copyright (C) 2023 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.android.devicelockcontroller.receivers;

import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;

import androidx.annotation.VisibleForTesting;

import com.android.devicelockcontroller.schedule.DeviceLockControllerScheduler;
import com.android.devicelockcontroller.schedule.DeviceLockControllerSchedulerProvider;
import com.android.devicelockcontroller.storage.GlobalParametersClient;
import com.android.devicelockcontroller.storage.UserParameters;
import com.android.devicelockcontroller.util.LogUtil;

import com.google.common.util.concurrent.FutureCallback;
import com.google.common.util.concurrent.Futures;
import com.google.common.util.concurrent.ListenableFuture;
import com.google.common.util.concurrent.MoreExecutors;

import java.util.concurrent.Executor;
import java.util.concurrent.Executors;

/**
 * Boot completed broadcast receiver to enqueue the check-in work for provision when device boots
 * for the first time.
 *
 * Only runs on system user and is disabled after check-in completes successfully.
 */
public final class CheckInBootCompletedReceiver extends BroadcastReceiver {

    private static final String TAG = "CheckInBootCompletedReceiver";
    private final Executor mExecutor;

    public CheckInBootCompletedReceiver() {
        mExecutor = Executors.newSingleThreadExecutor();
    }

    @VisibleForTesting
    CheckInBootCompletedReceiver(Executor executor) {
        mExecutor = executor;
    }

    @Override
    public void onReceive(Context context, Intent intent) {
        if (!intent.getAction().equals(Intent.ACTION_BOOT_COMPLETED)) return;

        LogUtil.i(TAG, "Received boot completed intent");

        if (!context.getUser().isSystem()) {
            throw new IllegalStateException(
                    "This receiver should not run on anything besides the system user");
        }
        Context applicationContext = context.getApplicationContext();
        DeviceLockControllerSchedulerProvider schedulerProvider =
                (DeviceLockControllerSchedulerProvider) applicationContext;
        DeviceLockControllerScheduler scheduler =
                schedulerProvider.getDeviceLockControllerScheduler();
        ListenableFuture<Boolean> needReschedule = Futures.transformAsync(
                Futures.submit(() -> UserParameters.needInitialCheckIn(context), mExecutor),
                needCheckIn -> {
                    if (needCheckIn) {
                        scheduler.scheduleInitialCheckInWork();
                        return Futures.immediateFuture(false);
                    } else {
                        return Futures.transform(
                                GlobalParametersClient.getInstance().isProvisionReady(),
                                ready -> !ready, MoreExecutors.directExecutor());
                    }
                }, mExecutor);
        Futures.addCallback(needReschedule,
                new FutureCallback<>() {
                    @Override
                    public void onSuccess(Boolean result) {
                        if (result) {
                            scheduler.notifyNeedRescheduleCheckIn();
                        }
                    }

                    @Override
                    public void onFailure(Throwable t) {
                        throw new RuntimeException(t);
                    }
                }, mExecutor);
    }
}
